#
# Copyright (c) 2021 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import sys
import importlib
import json

from unittest import TestCase, main
from pycg import utils

class TestBase(TestCase):
    snippet_dir = ""

    def setUp(self):
        def error():
            print ("Invalid module %s.%s" % (cg_mod, cg_class))
            print ("Set environment variables `CALL_GRAPH_CLASS` and `CALL_GRAPH_MODULE` properly")
            sys.exit(1)

        self.snippets_path = os.environ.get("SNIPPETS_PATH")
        cg_class = os.environ.get('CALL_GRAPH_CLASS', None)
        cg_mod = os.environ.get('CALL_GRAPH_MODULE', None)
        if not cg_class or not cg_mod:
            error()
        try:
            self.cg_mod = importlib.import_module(cg_mod)
        except ImportError:
            error()

        self.cg_class = getattr(self.cg_mod, cg_class)
        if not self.cg_class:
            error()

    def validate_snippet(self, snippet_path):
        output = self.get_snippet_output_cg(snippet_path)
        expected = self.get_snippet_expected_cg(snippet_path)

        self.assertEqual(output, expected)

    def get_snippet_path(self, name):
        return os.path.join(self.snippets_path, self.snippet_dir, name)

    def get_snippet_output_cg(self, snippet_path):
        main_path = os.path.join(snippet_path, "main.py")
        try:
            cg = self.cg_class([main_path], snippet_path, -1, utils.constants.KEY_ERR_OP)
            cg.analyze()
            return cg.output_key_errs()
        except Exception as e:
            cg.tearDown()
            raise e

    def get_snippet_expected_cg(self, snippet_path):
        cg_path = os.path.join(snippet_path, "key_errs.json")
        with open(cg_path, "r") as f:
            return json.loads(f.read())

    def assertEqual(self, actual, expected):
        super().assertEqual(actual, expected)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import ast

from pycg.processing.preprocessor import PreProcessor
from pycg.processing.postprocessor import PostProcessor
from pycg.processing.cgprocessor import CallGraphProcessor
from pycg.processing.keyerrprocessor import KeyErrProcessor

from pycg.machinery.scopes import ScopeManager
from pycg.machinery.definitions import DefinitionManager
from pycg.machinery.imports import ImportManager
from pycg.machinery.classes import ClassManager
from pycg.machinery.callgraph import CallGraph
from pycg.machinery.key_err import KeyErrors
from pycg.machinery.modules import ModuleManager
from pycg import utils

class CallGraphGenerator(object):
    def __init__(self, entry_points, package, max_iter, operation):
        self.entry_points = entry_points
        self.package = package
        self.state = None
        self.max_iter = max_iter
        self.operation = operation
        self.setUp()

    def setUp(self):
        self.import_manager = ImportManager()
        self.scope_manager = ScopeManager()
        self.def_manager = DefinitionManager()
        self.class_manager = ClassManager()
        self.module_manager = ModuleManager()
        self.cg = CallGraph()
        self.key_errs = KeyErrors()

    def extract_state(self):
        state = {}
        state["defs"] = {}
        for key, defi in self.def_manager.get_defs().items():
            state["defs"][key] = {
                "names": defi.get_name_pointer().get().copy(),
                "lit": defi.get_lit_pointer().get().copy()
            }

        state["scopes"] = {}
        for key, scope in self.scope_manager.get_scopes().items():
            state["scopes"][key] = set([x.get_ns() for (_, x) in scope.get_defs().items()])

        state["classes"] = {}
        for key, ch in self.class_manager.get_classes().items():
            state["classes"][key] = ch.get_mro().copy()
        return state

    def reset_counters(self):
        for key, scope in self.scope_manager.get_scopes().items():
            scope.reset_counters()

    def has_converged(self):
        if not self.state:
            return False

        curr_state = self.extract_state()

        # check defs
        for key, defi in curr_state["defs"].items():
            if not key in self.state["defs"]:
                return False
            if defi["names"] != self.state["defs"][key]["names"]:
                return False
            if defi["lit"] != self.state["defs"][key]["lit"]:
                return False

        # check scopes
        for key, scope in curr_state["scopes"].items():
            if not key in self.state["scopes"]:
                return False
            if scope != self.state["scopes"][key]:
                return False

        # check classes
        for key, ch in curr_state["classes"].items():
            if not key in self.state["classes"]:
                return False
            if ch != self.state["classes"][key]:
                return False

        return True

    def remove_import_hooks(self):
        self.import_manager.remove_hooks()

    def tearDown(self):
        self.remove_import_hooks()

    def _get_mod_name(self, entry, pkg):
        # We do this because we want __init__ modules to
        # only contain the parent module
        # since pycg can't differentiate between functions
        # coming from __init__ files.

        input_mod = utils.to_mod_name(
            os.path.relpath(entry, pkg))
        if input_mod.endswith("__init__"):
            input_mod = ".".join(input_mod.split(".")[:-1])

        return input_mod

    def do_pass(self, cls, install_hooks=False, *args, **kwargs):
        modules_analyzed = set()
        for entry_point in self.entry_points:
            input_pkg = self.package
            input_mod = self._get_mod_name(entry_point, input_pkg)
            input_file = os.path.abspath(entry_point)

            if not input_mod:
                continue

            if not input_pkg:
                input_pkg = os.path.dirname(input_file)

            if not input_mod in modules_analyzed:
                if install_hooks:
                    self.import_manager.set_pkg(input_pkg)
                    self.import_manager.install_hooks()

                processor = cls(input_file, input_mod,
                                modules_analyzed=modules_analyzed, *args, **kwargs)
                processor.analyze()
                modules_analyzed = modules_analyzed.union(processor.get_modules_analyzed())

                if install_hooks:
                    self.remove_import_hooks()

    def analyze(self):
        self.do_pass(PreProcessor, True,
                self.import_manager, self.scope_manager, self.def_manager,
                self.class_manager, self.module_manager)
        self.def_manager.complete_definitions()

        iter_cnt = 0
        while (self.max_iter < 0 or iter_cnt < self.max_iter) and (not self.has_converged()):
            self.state = self.extract_state()
            self.reset_counters()
            self.do_pass(PostProcessor, False,
                    self.import_manager, self.scope_manager, self.def_manager,
                    self.class_manager, self.module_manager)

            self.def_manager.complete_definitions()
            iter_cnt += 1

        self.reset_counters()
        if self.operation == utils.constants.CALL_GRAPH_OP:
            self.do_pass(CallGraphProcessor, False,
                    self.import_manager, self.scope_manager, self.def_manager,
                    self.class_manager, self.module_manager, call_graph=self.cg)
        elif self.operation == utils.constants.KEY_ERR_OP:
            self.do_pass(KeyErrProcessor, False,
                    self.import_manager, self.scope_manager, self.def_manager,
                    self.class_manager, self.key_errs)
        else:
            raise Exception("Invalid operation: " + self.operation)


    def output(self):
        return self.cg.get()

    def output_key_errs(self):
        return self.key_errs.get()

    def output_edges(self):
        return self.key_errors

    def output_edges(self):
        return self.cg.get_edges()

    def _generate_mods(self, mods):
        res = {}
        for mod, node in mods.items():
            res[mod] = {
                "filename": os.path.relpath(node.get_filename(), self.package)\
                    if node.get_filename() else None,
                "methods": node.get_methods()
            }
        return res

    def output_internal_mods(self):
        return self._generate_mods(self.module_manager.get_internal_modules())

    def output_external_mods(self):
        return self._generate_mods(self.module_manager.get_external_modules())

    def output_functions(self):
        functions = []
        for ns, defi in self.def_manager.get_defs().items():
            if defi.is_function_def():
                functions.append(ns)
        return functions

    def output_classes(self):
        classes = {}
        for cls, node in self.class_manager.get_classes().items():
            classes[cls] = {
                "mro": node.get_mro(),
                "module": node.get_module()
            }
        return classes

    def get_as_graph(self):
        return self.def_manager.get_defs().items()




import os

from base import TestBase

class AssignmentsTest(TestBase):
    snippet_dir = "assignments"

    def test_tuple(self):
        self.validate_snippet(self.get_snippet_path("tuple"))

    def test_chained(self):
        self.validate_snippet(self.get_snippet_path("chained"))

    def test_recursive_tuple(self):
        self.validate_snippet(self.get_snippet_path("recursive_tuple"))

    def test_starred(self):
        self.validate_snippet(self.get_snippet_path("starred"))




import os

from base import TestBase

class ClassesTest(TestBase):
    snippet_dir = "classes"

    def test_imported_call_without_init(self):
        self.validate_snippet(self.get_snippet_path("imported_call_without_init"))

    def test_direct_call(self):
        self.validate_snippet(self.get_snippet_path("direct_call"))

    def test_return_call_direct(self):
        self.validate_snippet(self.get_snippet_path("return_call_direct"))

    def test_nested_class_calls(self):
        self.validate_snippet(self.get_snippet_path("nested_class_calls"))

    def test_super_class_return(self):
        self.validate_snippet(self.get_snippet_path("super_class_return"))

    def test_base_class_calls_child(self):
        self.validate_snippet(self.get_snippet_path("base_class_calls_child"))

    def test_tuple_assignment(self):
        self.validate_snippet(self.get_snippet_path("tuple_assignment"))

    def test_return_call(self):
        self.validate_snippet(self.get_snippet_path("return_call"))

    def test_nested_call(self):
        self.validate_snippet(self.get_snippet_path("nested_call"))

    def test_imported_attr_access(self):
        self.validate_snippet(self.get_snippet_path("imported_attr_access"))

    def test_assigned_call(self):
        self.validate_snippet(self.get_snippet_path("assigned_call"))

    def test_self_assign_func(self):
        self.validate_snippet(self.get_snippet_path("self_assign_func"))

    def test_imported_call(self):
        self.validate_snippet(self.get_snippet_path("imported_call"))

    def test_base_class_attr(self):
        self.validate_snippet(self.get_snippet_path("base_class_attr"))

    def test_static_method_call(self):
        self.validate_snippet(self.get_snippet_path("static_method_call"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))

    def test_instance(self):
        self.validate_snippet(self.get_snippet_path("instance"))

    def test_imported_nested_attr_access(self):
        self.validate_snippet(self.get_snippet_path("imported_nested_attr_access"))

    def test_parameter_call(self):
        self.validate_snippet(self.get_snippet_path("parameter_call"))

    def test_self_assignment(self):
        self.validate_snippet(self.get_snippet_path("self_assignment"))

    def test_self_call(self):
        self.validate_snippet(self.get_snippet_path("self_call"))

    def test_assigned_self_call(self):
        self.validate_snippet(self.get_snippet_path("assigned_self_call"))




import os

from base import TestBase

class MroTest(TestBase):
    snippet_dir = "mro"

    def test_basic_init(self):
        self.validate_snippet(self.get_snippet_path("basic_init"))

    def test_parents_same_superclass(self):
        self.validate_snippet(self.get_snippet_path("parents_same_superclass"))

    def test_basic(self):
        self.validate_snippet(self.get_snippet_path("basic"))

    def test_two_parents(self):
        self.validate_snippet(self.get_snippet_path("two_parents"))

    def test_super_call(self):
        self.validate_snippet(self.get_snippet_path("super_call"))

    def test_self_assignment(self):
        self.validate_snippet(self.get_snippet_path("self_assignment"))

    def test_two_parents_method_defined(self):
        self.validate_snippet(self.get_snippet_path("two_parents_method_defined"))




import os

from base import TestBase

class ReturnsTest(TestBase):
    snippet_dir = "returns"

    def test_return_complex(self):
        self.validate_snippet(self.get_snippet_path("return_complex"))

    def test_imported_call(self):
        self.validate_snippet(self.get_snippet_path("imported_call"))

    def test_nested_import_call(self):
        self.validate_snippet(self.get_snippet_path("nested_import_call"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))




import os

from base import TestBase

class DirectCallsTest(TestBase):
    snippet_dir = "direct_calls"

    def test_with_parameters(self):
        self.validate_snippet(self.get_snippet_path("with_parameters"))

    def test_imported_return_call(self):
        self.validate_snippet(self.get_snippet_path("imported_return_call"))

    def test_return_call(self):
        self.validate_snippet(self.get_snippet_path("return_call"))

    def test_assigned_call(self):
        self.validate_snippet(self.get_snippet_path("assigned_call"))



#!/usr/bin/env python3

import os

FILE_DIR = os.path.dirname(os.path.relpath(__file__))
SNIPPETS_DIR = "snippets"

base_template = """
import os

from base import TestBase

class {cls}Test(TestBase):
    snippet_dir = "{dir}"
"""

test_template = """
    def test_{name}(self):
        self.validate_snippet(self.get_snippet_path("{name}"))
"""

def create_test_case(name):
    test_name = name + "_test.py"
    # TODO: capitalize
    capitalized = "".join([x.title() for x in name.split("_")])
    template = base_template.format(
        cls=capitalized,
        dir=name
    )
    for name in os.listdir(os.path.join(FILE_DIR, SNIPPETS_DIR, name)):
        if name == "." or name == "..":
            continue
        template += test_template.format(name=name)

    with open(os.path.join(FILE_DIR, test_name), "w+") as f:
        f.write(template)

for name in os.listdir(os.path.join(FILE_DIR, SNIPPETS_DIR)):
    if name == "." or name == "..":
        continue

    create_test_case(name)




import os

from base import TestBase

class ImportsTest(TestBase):
    snippet_dir = "imports"

    def test_relative_import_with_name(self):
        self.validate_snippet(self.get_snippet_path("relative_import_with_name"))

    def test_submodule_import_from(self):
        self.validate_snippet(self.get_snippet_path("submodule_import_from"))

    def test_import_as(self):
        self.validate_snippet(self.get_snippet_path("import_as"))

    def test_submodule_import_all(self):
        self.validate_snippet(self.get_snippet_path("submodule_import_all"))

    def test_parent_import(self):
        self.validate_snippet(self.get_snippet_path("parent_import"))

    def test_chained_import(self):
        self.validate_snippet(self.get_snippet_path("chained_import"))

    def test_init_func_import(self):
        self.validate_snippet(self.get_snippet_path("init_func_import"))

    def test_submodule_import_as(self):
        self.validate_snippet(self.get_snippet_path("submodule_import_as"))

    def test_relative_import(self):
        self.validate_snippet(self.get_snippet_path("relative_import"))

    def test_import_all(self):
        self.validate_snippet(self.get_snippet_path("import_all"))

    def test_simple_import(self):
        self.validate_snippet(self.get_snippet_path("simple_import"))

    def test_init_import(self):
        self.validate_snippet(self.get_snippet_path("init_import"))

    def test_submodule_import(self):
        self.validate_snippet(self.get_snippet_path("submodule_import"))

    def test_import_from(self):
        self.validate_snippet(self.get_snippet_path("import_from"))




import os

from base import TestBase

class FunctionsTest(TestBase):
    snippet_dir = "functions"

    def test_assigned_call(self):
        self.validate_snippet(self.get_snippet_path("assigned_call"))

    def test_imported_call(self):
        self.validate_snippet(self.get_snippet_path("imported_call"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))

    def test_assigned_call_lit_param(self):
        self.validate_snippet(self.get_snippet_path("assigned_call_lit_param"))



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import ast

from pycg.processing.base import ProcessingBase
from pycg.machinery.definitions import Definition
from pycg import utils

class PostProcessor(ProcessingBase):
    def __init__(self, input_file, modname, import_manager,
            scope_manager, def_manager, class_manager, module_manager, modules_analyzed=None):
        super().__init__(input_file, modname, modules_analyzed)
        self.import_manager = import_manager
        self.scope_manager = scope_manager
        self.def_manager = def_manager
        self.class_manager = class_manager
        self.module_manager = module_manager
        self.closured = self.def_manager.transitive_closure()

    def visit_Lambda(self, node):
        counter = self.scope_manager.get_scope(self.current_ns).inc_lambda_counter()
        lambda_name = utils.get_lambda_name(counter)
        super().visit_Lambda(node, lambda_name)

    def visit_Call(self, node):
        self.visit(node.func)

        names = self.retrieve_call_names(node)
        if not names:
            return

        self.last_called_names = names

        for name in names:
            defi = self.def_manager.get(name)
            if not defi:
                continue
            if defi.get_type() == utils.constants.CLS_DEF:
                self.update_parent_classes(defi)
                defi = self.def_manager.get(utils.join_ns(defi.get_ns(), utils.constants.CLS_INIT))
                if not defi:
                    continue
            self.iterate_call_args(defi, node)

    def visit_Assign(self, node):
        self._visit_assign(node.value, node.targets)

    def visit_Return(self, node):
        self._visit_return(node)

    def visit_Yield(self, node):
        self._visit_return(node)

    def visit_For(self, node):
        # only handle name targets
        if isinstance(node.target, ast.Name):
            target_def = self.def_manager.get(utils.join_ns(self.current_ns, node.target.id))
            # if the target definition exists
            if target_def:
                iter_decoded = self.decode_node(node.iter)
                # assign the target to the return value
                # of the next function
                for item in iter_decoded:
                    if not isinstance(item, Definition):
                        continue
                    # return value for generators
                    for name in self.closured.get(item.get_ns(), []):
                        # If there exists a next method on the iterable
                        # and if yes, add a pointer to it
                        next_defi = self.def_manager.get(utils.join_ns(name,
                            utils.constants.NEXT_METHOD, utils.constants.RETURN_NAME))
                        if next_defi:
                            for name in self.closured.get(next_defi.get_ns(), []):
                                target_def.get_name_pointer().add(name)
                        else: # otherwise, add a pointer to the name (e.g. a yield)
                            target_def.get_name_pointer().add(name)

        super().visit_For(node)

    def visit_Return(self, node):
        self._visit_return(node)

    def visit_Yield(self, node):
        self._visit_return(node)

    def visit_AsyncFunctionDef(self, node):
        self.visit_FunctionDef(node)

    def visit_FunctionDef(self, node):
        # here we iterate decorators
        if node.decorator_list:
            fn_def = self.def_manager.get(utils.join_ns(self.current_ns, node.name))
            reversed_decorators = list(reversed(node.decorator_list))

            # add to the name pointer of the function definition
            # the return value of the first decorator
            # since, now the function is a namespace to that point
            if hasattr(fn_def, "decorator_names") and reversed_decorators:
                last_decoded = self.decode_node(reversed_decorators[-1])
                for d in last_decoded:
                    if not isinstance(d, Definition):
                        continue
                    fn_def.decorator_names.add(utils.join_ns(d.get_ns(), utils.constants.RETURN_NAME))

            previous_names = self.closured.get(fn_def.get_ns(), set())
            for decorator in reversed_decorators:
                # assign the previous_def as the first parameter of the decorator
                decoded = self.decode_node(decorator)
                new_previous_names = set()
                for d in decoded:
                    if not isinstance(d, Definition):
                        continue
                    for name in self.closured.get(d.get_ns(), []):
                        return_ns = utils.join_ns(name, utils.constants.RETURN_NAME)

                        if self.closured.get(return_ns, None) == None:
                            continue

                        new_previous_names = new_previous_names.union(self.closured.get(return_ns))

                        for prev_name in previous_names:
                            pos_arg_names = d.get_name_pointer().get_pos_arg(0)
                            if not pos_arg_names:
                                continue
                            for name in pos_arg_names:
                                arg_def = self.def_manager.get(name)
                                arg_def.get_name_pointer().add(prev_name)
                previous_names = new_previous_names

        super().visit_FunctionDef(node)

    def visit_ClassDef(self, node):
        # create a definition for the class (node.name)
        cls_def = self.def_manager.handle_class_def(self.current_ns, node.name)

        # iterate bases to compute MRO for the class
        cls = self.class_manager.get(cls_def.get_ns())
        if not cls:
            cls = self.class_manager.create(cls_def.get_ns(), self.modname)

        cls.clear_mro()
        for base in node.bases:
            # all bases are of the type ast.Name
            self.visit(base)

            bases = self.decode_node(base)
            for base_def in bases:
                if not isinstance(base_def, Definition):
                    continue
                names = set()
                if base_def.get_name_pointer().get():
                    names = base_def.get_name_pointer().get()
                else:
                    names.add(base_def.get_ns())
                for name in names:
                    # add the base as a parent
                    cls.add_parent(name)

                    # add the base's parents
                    parent_cls = self.class_manager.get(name)
                    if parent_cls:
                        cls.add_parent(parent_cls.get_mro())

        cls.compute_mro()

        super().visit_ClassDef(node)

    def visit_List(self, node):
        # Works similarly with dicts
        current_scope = self.scope_manager.get_scope(self.current_ns)
        list_counter = current_scope.inc_list_counter()
        list_name = utils.get_list_name(list_counter)
        list_full_ns = utils.join_ns(self.current_ns, list_name)

        # create a scope for the list
        list_scope = self.scope_manager.create_scope(list_full_ns, current_scope)

        # create a list definition
        list_def = self.def_manager.get(list_full_ns)
        if not list_def:
            list_def = self.def_manager.create(list_full_ns, utils.constants.NAME_DEF)
        current_scope.add_def(list_name, list_def)

        self.name_stack.append(list_name)
        for idx, elt in enumerate(node.elts):
            self.visit(elt)
            key_full_ns = utils.join_ns(list_def.get_ns(), utils.get_int_name(idx))
            key_def = self.def_manager.get(key_full_ns)
            if not key_def:
                key_def = self.def_manager.create(key_full_ns, utils.constants.NAME_DEF)

            decoded_elt = self.decode_node(elt)
            for v in decoded_elt:
                if isinstance(v, Definition):
                    key_def.get_name_pointer().add(v.get_ns())
                else:
                    key_def.get_lit_pointer().add(v)

        self.name_stack.pop()

    def visit_Dict(self, node):
        # 1. create a scope using a counter
        # 2. Iterate keys and add them as children of the scope
        # 3. Iterate values and makes a points to connection with the keys
        current_scope = self.scope_manager.get_scope(self.current_ns)
        dict_counter = current_scope.inc_dict_counter()
        dict_name = utils.get_dict_name(dict_counter)
        dict_full_ns = utils.join_ns(self.current_ns, dict_name)

        # create a scope for the dict
        dict_scope = self.scope_manager.create_scope(dict_full_ns, current_scope)

        # Create a dict definition
        dict_def = self.def_manager.get(dict_full_ns)
        if not dict_def:
            dict_def = self.def_manager.create(dict_full_ns, utils.constants.NAME_DEF)
        # add it to the current scope
        current_scope.add_def(dict_name, dict_def)

        self.name_stack.append(dict_name)
        for key, value in zip(node.keys, node.values):
            if key:
                self.visit(key)
            if value:
                self.visit(value)
            decoded_key = self.decode_node(key)
            decoded_value = self.decode_node(value)

            # iterate decoded keys and values
            # to do the assignment operation
            for k in decoded_key:
                if isinstance(k, Definition):
                    # get literal pointer
                    names = k.get_lit_pointer().get()
                else:
                    names = set()
                    if isinstance(k, list):
                        continue
                    names.add(k)
                for name in names:
                    # create a definition for the key
                    if isinstance(name, int):
                        name = utils.get_int_name(name)
                    key_full_ns = utils.join_ns(dict_def.get_ns(), str(name))
                    key_def = self.def_manager.get(key_full_ns)
                    if not key_def:
                        key_def = self.def_manager.create(key_full_ns, utils.constants.NAME_DEF)
                    dict_scope.add_def(str(name), key_def)
                    for v in decoded_value:
                        if isinstance(v, Definition):
                            key_def.get_name_pointer().add(v.get_ns())
                        else:
                            key_def.get_lit_pointer().add(v)
        self.name_stack.pop()

    def update_parent_classes(self, defi):
        cls = self.class_manager.get(defi.get_ns())
        if not cls:
            return
        current_scope = self.scope_manager.get_scope(defi.get_ns())
        for parent in cls.get_mro():
            parent_def = self.def_manager.get(parent)
            if not parent_def:
                continue
            parent_scope = self.scope_manager.get_scope(parent)
            if not parent_scope:
                continue
            parent_items = list(parent_scope.get_defs().keys())
            for key, child_def in current_scope.get_defs().items():
                if key == "__init__":
                    continue
                # resolve name from the parent_def
                names = self.find_cls_fun_ns(parent_def.get_ns(), key)

                new_ns = utils.join_ns(parent_def.get_ns(), key)
                new_def = self.def_manager.get(new_ns)
                if not new_def:
                    new_def = self.def_manager.create(new_ns, utils.constants.NAME_DEF)

                new_def.get_name_pointer().add_set(names)
                new_def.get_name_pointer().add(child_def.get_ns())

    def analyze_submodules(self):
        super().analyze_submodules(PostProcessor, self.import_manager,
                self.scope_manager, self.def_manager, self.class_manager,
                self.module_manager, modules_analyzed=self.get_modules_analyzed())

    def analyze(self):
        self.visit(ast.parse(self.contents, self.filename))
        self.analyze_submodules()



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import symtable
from pycg import utils

class ScopeManager(object):
    """Manages the scope entries"""

    def __init__(self):
        self.scopes = {}

    def handle_module(self, modulename, filename, contents):
        functions = []
        classes = []
        def process(namespace, parent, table):
            if table.get_name() == 'top' and table.get_lineno() == 0:
                name = ''
            else:
                name = table.get_name()

            if name:
                fullns = utils.join_ns(namespace, name)
            else:
                fullns = namespace

            if table.get_type() == "function":
                functions.append(fullns)

            if table.get_type() == "class":
                classes.append(fullns)

            sc = self.create_scope(fullns, parent)

            for t in table.get_children():
                process(fullns, sc, t)

        process(modulename, None, symtable.symtable(contents, filename, compile_type="exec"))
        return {"functions": functions, "classes": classes}

    def handle_assign(self, ns, target, defi):
        scope = self.get_scope(ns)
        if scope:
            scope.add_def(target, defi)

    def get_def(self, current_ns, var_name):
        current_scope = self.get_scope(current_ns)
        while current_scope:
            defi = current_scope.get_def(var_name)
            if defi:
                return defi
            current_scope = current_scope.parent

    def get_scope(self, namespace):
        if namespace in self.get_scopes():
            return self.get_scopes()[namespace]

    def create_scope(self, namespace, parent):
        if not namespace in self.scopes:
            sc = ScopeItem(namespace, parent)
            self.scopes[namespace] = sc
        return self.scopes[namespace]

    def get_scopes(self):
        return self.scopes

class ScopeItem(object):
    def __init__(self, fullns, parent):
        if parent and not isinstance(parent, ScopeItem):
            raise ScopeError("Parent must be a ScopeItem instance")

        if not isinstance(fullns, str):
            raise ScopeError("Namespace should be a string")

        self.parent = parent
        self.defs = {}
        self.lambda_counter = 0
        self.dict_counter = 0
        self.list_counter = 0
        self.fullns = fullns

    def get_ns(self):
        return self.fullns

    def get_defs(self):
        return self.defs

    def get_def(self, name):
        defs = self.get_defs()
        if name in defs:
            return defs[name]

    def get_lambda_counter(self):
        return self.lambda_counter

    def get_dict_counter(self):
        return self.dict_counter

    def get_list_counter(self):
        return self.list_counter

    def inc_lambda_counter(self, val=1):
        self.lambda_counter += val
        return self.lambda_counter

    def inc_dict_counter(self, val=1):
        self.dict_counter += val
        return self.dict_counter

    def inc_list_counter(self, val=1):
        self.list_counter += val
        return self.list_counter

    def reset_counters(self):
        self.lambda_counter = 0
        self.dict_counter = 0
        self.list_counter = 0

    def add_def(self, name, defi):
        self.defs[name] = defi

    def merge_def(self, name, to_merge):
        if not name in self.defs:
            self.defs[name] = to_merge
            return

        self.defs[name].merge_points_to(to_merge.get_points_to())

class ScopeError(Exception):
    pass




import os

from base import TestBase

class DecoratorsTest(TestBase):
    snippet_dir = "decorators"

    def test_return_different_func(self):
        self.validate_snippet(self.get_snippet_path("return_different_func"))

    def test_param_call(self):
        self.validate_snippet(self.get_snippet_path("param_call"))

    def test_nested_decorators(self):
        self.validate_snippet(self.get_snippet_path("nested_decorators"))

    def test_assigned(self):
        self.validate_snippet(self.get_snippet_path("assigned"))

    def test_return(self):
        self.validate_snippet(self.get_snippet_path("return"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))

    def test_nested(self):
        self.validate_snippet(self.get_snippet_path("nested"))



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from pycg.machinery.pointers import NamePointer, LiteralPointer
from pycg import utils

class DefinitionManager(object):
    def __init__(self):
        self.defs = {}

    def create(self, ns, def_type):
        if not ns or not isinstance(ns, str):
            raise DefinitionError("Invalid namespace argument")
        if not def_type in Definition.types:
            raise DefinitionError("Invalid def type argument")
        if self.get(ns):
            raise DefinitionError("Definition already exists")

        self.defs[ns] = Definition(ns, def_type)
        return self.defs[ns]

    def assign(self, ns, defi):
        self.defs[ns] = Definition(ns, defi.get_type())
        self.defs[ns].merge(defi)

        # if it is a function def, we need to create a return pointer
        if defi.is_function_def():
            return_ns = utils.join_ns(ns, utils.constants.RETURN_NAME)
            self.defs[return_ns] = Definition(return_ns, utils.constants.NAME_DEF)
            self.defs[return_ns].get_name_pointer().add(
                utils.join_ns(defi.get_ns(), utils.constants.RETURN_NAME))

        return self.defs[ns]

    def get(self, ns):
        if ns in self.defs:
            return self.defs[ns]

    def get_defs(self):
        return self.defs

    def handle_function_def(self, parent_ns, fn_name):
        full_ns = utils.join_ns(parent_ns, fn_name)
        defi = self.get(full_ns)
        if not defi:
            defi = self.create(full_ns, utils.constants.FUN_DEF)
            defi.decorator_names = set()

        return_ns = utils.join_ns(full_ns, utils.constants.RETURN_NAME)
        if not self.get(return_ns):
            self.create(return_ns, utils.constants.NAME_DEF)

        return defi

    def handle_class_def(self, parent_ns, cls_name):
        full_ns = utils.join_ns(parent_ns, cls_name)
        defi = self.get(full_ns)
        if not defi:
            defi = self.create(full_ns, utils.constants.CLS_DEF)

        return defi

    def transitive_closure(self):
        closured = {}
        def dfs(defi):
            name_pointer = defi.get_name_pointer()
            new_set = set()
            # bottom
            if not closured.get(defi.get_ns(), None) == None:
                return closured[defi.get_ns()]

            if not name_pointer.get():
                new_set.add(defi.get_ns())

            closured[defi.get_ns()] = new_set

            for name in name_pointer.get():
                if not self.defs.get(name, None):
                    continue
                items = dfs(self.defs[name])
                if not items:
                    items = set([name])
                new_set = new_set.union(items)

            closured[defi.get_ns()] = new_set
            return closured[defi.get_ns()]

        for ns, current_def in self.defs.items():
            if closured.get(current_def, None) == None:
                dfs(current_def)

        return closured

    def complete_definitions(self):
        # THE MOST expensive part of this tool's process
        # TODO: IMPROVE COMPLEXITY
        def update_pointsto_args(pointsto_args, arg, name):
            changed_something = False
            if arg == pointsto_args:
                return False
            for pointsto_arg in pointsto_args:
                if not self.defs.get(pointsto_arg, None):
                    continue
                if pointsto_arg == name:
                    continue
                pointsto_arg_def = self.defs[pointsto_arg].get_name_pointer()
                if pointsto_arg_def == pointsto_args:
                    continue

                # sometimes we may end up with a cycle
                if pointsto_arg in arg:
                    arg.remove(pointsto_arg)

                for item in arg:
                    if not item in pointsto_arg_def.get():
                        if self.defs.get(item, None) != None:
                            changed_something = True
                    # HACK: this check shouldn't be needed
                    # if we remove this the following breaks:
                    # x = lambda x: x + 1
                    # x(1)
                    # since on line 184 we don't discriminate between literal values and name values
                    if not self.defs.get(item, None):
                        continue
                    pointsto_arg_def.add(item)
            return changed_something

        for i in range(len(self.defs)):
            changed_something = False
            for ns, current_def in self.defs.items():
                # the name pointer of the definition we're currently iterating
                current_name_pointer = current_def.get_name_pointer()
                # iterate the names the current definition points to items
                # for name in current_name_pointer.get():
                for name in current_name_pointer.get().copy():

                    # get the name pointer of the points to name
                    if not self.defs.get(name, None):
                        continue
                    if name == ns:
                        continue

                    pointsto_name_pointer = self.defs[name].get_name_pointer()
                    # iterate the arguments of the definition we're currently iterating
                    for arg_name, arg in current_name_pointer.get_args().items():
                        pos = current_name_pointer.get_pos_of_name(arg_name)
                        if not pos is None:
                            pointsto_args = pointsto_name_pointer.get_pos_arg(pos)
                            if not pointsto_args:
                                pointsto_name_pointer.add_pos_arg(pos, None, arg)
                                continue
                        else:
                            pointsto_args = pointsto_name_pointer.get_arg(arg_name)
                            if not pointsto_args:
                                pointsto_name_pointer.add_arg(arg_name, arg)
                                continue
                        changed_something = changed_something or update_pointsto_args(pointsto_args, arg, current_def.get_ns())

            if not changed_something:
                break


class Definition(object):
    types = [
        utils.constants.FUN_DEF,
        utils.constants.MOD_DEF,
        utils.constants.NAME_DEF,
        utils.constants.CLS_DEF,
        utils.constants.EXT_DEF
    ]

    def __init__(self, fullns, def_type):
        self.fullns = fullns
        self.points_to = {
            "lit": LiteralPointer(),
            "name": NamePointer()
        }
        self.def_type = def_type

    def get_type(self):
        return self.def_type

    def is_function_def(self):
        return self.def_type == utils.constants.FUN_DEF

    def is_ext_def(self):
        return self.def_type == utils.constants.EXT_DEF

    def is_callable(self):
        return (self.is_function_def() or self.is_ext_def())

    def get_lit_pointer(self):
        return self.points_to["lit"]

    def get_name_pointer(self):
        return self.points_to["name"]

    def get_name(self):
        return self.fullns.split(".")[-1]

    def get_ns(self):
        return self.fullns

    def merge(self, to_merge):
        for name, pointer in to_merge.points_to.items():
            self.points_to[name].merge(pointer)

class DefinitionError(Exception):
    pass




import os

from base import TestBase

class ArgsTest(TestBase):
    snippet_dir = "args"

    def test_nested_call(self):
        self.validate_snippet(self.get_snippet_path("nested_call"))

    def test_param_call(self):
        self.validate_snippet(self.get_snippet_path("param_call"))

    def test_assigned_call(self):
        self.validate_snippet(self.get_snippet_path("assigned_call"))

    def test_imported_call(self):
        self.validate_snippet(self.get_snippet_path("imported_call"))

    def test_imported_assigned_call(self):
        self.validate_snippet(self.get_snippet_path("imported_assigned_call"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))




import os

from base import TestBase

class KwargsTest(TestBase):
    snippet_dir = "kwargs"

    def test_chained_call(self):
        self.validate_snippet(self.get_snippet_path("chained_call"))

    def test_assigned_call(self):
        self.validate_snippet(self.get_snippet_path("assigned_call"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))



import os
import sys
import json
import argparse

from pycg.pycg import CallGraphGenerator
from pycg import formats
from pycg.utils.constants import CALL_GRAPH_OP, KEY_ERR_OP

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("entry_point",
        nargs="*",
        help="Entry points to be processed")
    parser.add_argument(
        "--package",
        help="Package containing the code to be analyzed",
        default=None
    )
    parser.add_argument(
        "--fasten",
        help="Produce call graph using the FASTEN format",
        action="store_true",
        default=False
    )
    parser.add_argument(
        "--product",
        help="Package name",
        default=""
    )
    parser.add_argument(
        "--forge",
        help="Source the product was downloaded from",
        default=""
    )
    parser.add_argument(
        "--version",
        help="Version of the product",
        default=""
    )
    parser.add_argument(
        "--timestamp",
        help="Timestamp of the package's version",
        default=0
    )
    parser.add_argument(
        "--max-iter",
        type=int,
        help=("Maximum number of iterations through source code. " +
            "If not specified a fix-point iteration will be performed."),
        default=-1
    )
    parser.add_argument(
        '--operation',
        type=str,
        choices=[CALL_GRAPH_OP, KEY_ERR_OP],
        help=("Operation to perform. " +
             "Choose " + CALL_GRAPH_OP + " for call graph generation (default)" +
             " or " + KEY_ERR_OP + " for key error detection on dictionaries."),
        default=CALL_GRAPH_OP
    )

    parser.add_argument(
        "--as-graph-output",
        help="Output for the assignment graph",
        default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Output path",
        default=None
    )

    args = parser.parse_args()

    cg = CallGraphGenerator(args.entry_point, args.package,
                        args.max_iter, args.operation)
    cg.analyze()

    if args.operation == CALL_GRAPH_OP:
        if args.fasten:
            formatter = formats.Fasten(cg, args.package,
                args.product, args.forge, args.version, args.timestamp)
        else:
            formatter = formats.Simple(cg)
        output = formatter.generate()
    else:
        output = cg.output_key_errs()

    as_formatter = formats.AsGraph(cg)

    if args.output:
        with open(args.output, "w+") as f:
            f.write(json.dumps(output))
    else:
        print (json.dumps(output))

    if args.as_graph_output:
        with open(args.as_graph_output, "w+") as f:
            f.write(json.dumps(as_formatter.generate()))

if __name__ == "__main__":
    main()




import os

from base import TestBase

class DictsTest(TestBase):
    snippet_dir = "dicts"

    def test_type_coercion(self):
        self.validate_snippet(self.get_snippet_path("type_coercion"))

    def test_update(self):
        self.validate_snippet(self.get_snippet_path("update"))

    def test_ext_key(self):
        self.validate_snippet(self.get_snippet_path("ext_key"))

    def test_return(self):
        self.validate_snippet(self.get_snippet_path("return"))

    def test_param(self):
        self.validate_snippet(self.get_snippet_path("param"))

    def test_return_assign(self):
        self.validate_snippet(self.get_snippet_path("return_assign"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))

    def test_assign(self):
        self.validate_snippet(self.get_snippet_path("assign"))

    def test_param_key(self):
        self.validate_snippet(self.get_snippet_path("param_key"))

    def test_new_key_param(self):
        self.validate_snippet(self.get_snippet_path("new_key_param"))

    def test_nested(self):
        self.validate_snippet(self.get_snippet_path("nested"))

    def test_add_key(self):
        self.validate_snippet(self.get_snippet_path("add_key"))




import os

from base import TestBase

class ListsTest(TestBase):
    snippet_dir = "lists"

    def test_comprehension_val(self):
        self.validate_snippet(self.get_snippet_path("comprehension_val"))

    def test_slice(self):
        self.validate_snippet(self.get_snippet_path("slice"))

    def test_simple(self):
        self.validate_snippet(self.get_snippet_path("simple"))

    def test_comprehension_if(self):
        self.validate_snippet(self.get_snippet_path("comprehension_if"))

    def test_nested_comprehension(self):
        self.validate_snippet(self.get_snippet_path("nested_comprehension"))

    def test_param_index(self):
        self.validate_snippet(self.get_snippet_path("param_index"))

    def test_nested(self):
        self.validate_snippet(self.get_snippet_path("nested"))

    def test_ext_index(self):
        self.validate_snippet(self.get_snippet_path("ext_index"))



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
class CallGraph(object):
    def __init__(self):
        self.cg = {}
        self.modnames = {}

    def add_node(self, name, modname=""):
        if not isinstance(name, str):
            raise CallGraphError("Only string node names allowed")
        if not name:
            raise CallGraphError("Empty node name")

        if not name in self.cg:
            self.cg[name] = set()
            self.modnames[name] = modname

        if name in self.cg and not self.modnames[name]:
            self.modnames[name] = modname

    def add_edge(self, src, dest):
        self.add_node(src)
        self.add_node(dest)
        self.cg[src].add(dest)

    def get(self):
        return self.cg

    def get_edges(self):
        output = []
        for src in self.cg:
            for dst in self.cg[src]:
                output.append([src, dst])
        return output

    def get_modules(self):
        return self.modnames


class CallGraphError(Exception):
    pass




import os

from base import TestBase

class ExternalTest(TestBase):
    snippet_dir = "external"

    def test_attribute(self):
        self.validate_snippet(self.get_snippet_path("attribute"))

    def test_function_assigned(self):
        self.validate_snippet(self.get_snippet_path("function_assigned"))

    def test_attribute_assigned(self):
        self.validate_snippet(self.get_snippet_path("attribute_assigned"))

    def test_function_asname(self):
        self.validate_snippet(self.get_snippet_path("function_asname"))

    def test_function(self):
        self.validate_snippet(self.get_snippet_path("function"))

    def test_cls_parent(self):
        self.validate_snippet(self.get_snippet_path("cls_parent"))




import os

from base import TestBase

class GeneratorsTest(TestBase):
    snippet_dir = "generators"

    def test_iter_param(self):
        self.validate_snippet(self.get_snippet_path("iter_param"))

    def test_no_iter(self):
        self.validate_snippet(self.get_snippet_path("no_iter"))

    def test_iter_return(self):
        self.validate_snippet(self.get_snippet_path("iter_return"))

    def test_iterable(self):
        self.validate_snippet(self.get_snippet_path("iterable"))

    def test_iterable_assigned(self):
        self.validate_snippet(self.get_snippet_path("iterable_assigned"))

    def test_yield(self):
        self.validate_snippet(self.get_snippet_path("yield"))




import os

from base import TestBase

class ExceptionsTest(TestBase):
    snippet_dir = "exceptions"

    def test_raise(self):
        self.validate_snippet(self.get_snippet_path("raise"))

    def test_raise_attr(self):
        self.validate_snippet(self.get_snippet_path("raise_attr"))

    def test_raise_assigned(self):
        self.validate_snippet(self.get_snippet_path("raise_assigned"))



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os

from setuptools import setup, find_packages
from subprocess import call

def get_long_desc():
    with open("README.md", "r") as readme:
        desc = readme.read()

    return desc

def setup_package():
    setup(
        name='pycg',
        version='0.0.6',
        description='Practical Python Call Graphs',
        long_description=get_long_desc(),
        long_description_content_type="text/markdown",
        url='https://github.com/vitsalis/pycg',
        license='Apache Software License',
        packages=find_packages(),
        install_requires=[],
        python_requires='>=3.4',
        entry_points = {
            'console_scripts': [
                'pycg=pycg.__main__:main',
            ],
        },
        classifiers=[
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3'
        ],
        author = 'Vitalis Salis',
        author_email = 'vitsalis@gmail.com'
    )

if __name__ == '__main__':
    setup_package()




import os

from base import TestBase

class BuiltinsTest(TestBase):
    snippet_dir = "builtins"

    def test_types(self):
        self.validate_snippet(self.get_snippet_path("types"))

    def test_map(self):
        self.validate_snippet(self.get_snippet_path("map"))

    def test_functions(self):
        self.validate_snippet(self.get_snippet_path("functions"))



#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
class Pointer(object):
    def __init__(self):
        self.values = set()

    def add(self, item):
        self.values.add(item)

    def add_set(self, s):
        self.values = self.values.union(s)

    def get(self):
        return self.values

    def merge(self, pointer):
        self.values = self.values.union(pointer.values)

class LiteralPointer(Pointer):
    STR_LIT = "STRING"
    INT_LIT = "INTEGER"
    UNK_LIT = "UNKNOWN"

    # no need to add the actual item
    def add(self, item):
        if isinstance(item, str):
            self.values.add(item)
        elif isinstance(item, int):
            self.values.add(item)
        else:
            self.values.add(self.UNK_LIT)

class NamePointer(Pointer):
    def __init__(self):
        super().__init__()
        self.pos_to_name = {}
        self.name_to_pos = {}
        self.args = {}

    def _sanitize_pos(self, pos):
        try:
            int(pos)
        except ValueError:
            raise PointerError("Invalid position for argument")

        return pos

    def get_or_create(self, name):
        if not name in self.args:
            self.args[name] = set()
        return self.args[name]

    def add_arg(self, name, item):
        arg = self.get_or_create(name)
        if isinstance(item, str):
            self.args[name].add(item)
        elif isinstance(item, set):
            self.args[name] = self.args[name].union(item)
        else:
            raise Exception()

    def add_lit_arg(self, name, item):
        arg = self.get_or_create(name)
        if isinstance(item, str):
            arg.add(LiteralPointer.STR_LIT)
        elif isinstance(item, int):
            arg.add(LiteralPointer.INT_LIT)
        else:
            arg.add(LiteralPointer.UNK_LIT)

    def add_pos_arg(self, pos, name, item):
        pos = self._sanitize_pos(pos)
        if not name:
            if self.pos_to_name.get(pos, None):
                name = self.pos_to_name[pos]
            else:
                name = str(pos)
        self.pos_to_name[pos] = name
        self.name_to_pos[name] = pos

        self.add_arg(name, item)

    def add_name_arg(self, name, item):
        self.add_arg(name, item)

    def add_pos_lit_arg(self, pos, name, item):
        pos = self._sanitize_pos(pos)
        if not name:
            name = str(pos)
        self.pos_to_name[pos] = name
        self.name_to_pos[name] = pos
        self.add_lit_arg(name, item)

    def get_pos_arg(self, pos):
        pos = self._sanitize_pos(pos)
        name = self.pos_to_name.get(pos, None)
        return self.get_arg(name)

    def get_arg(self, name):
        if self.args.get(name, None):
            return self.args[name]

    def get_args(self):
        return self.args

    def get_pos_args(self):
        args = {}
        for pos, name in self.pos_to_name.items():
            args[pos] = self.args[name]
        return args

    def get_pos_of_name(self, name):
        if name in self.name_to_pos:
            return self.name_to_pos[name]

    def get_pos_names(self):
        return self.pos_to_name

    def merge(self, pointer):
        super().merge(pointer)
        if hasattr(pointer, "get_pos_names"):
            for pos, name in pointer.get_pos_names().items():
                self.pos_to_name[pos] = name
            for name, arg in pointer.get_args().items():
                self.add_arg(name, arg)

class PointerError(Exception):
    pass




import os

from base import TestBase

class LambdasTest(TestBase):
    snippet_dir = "lambdas"

    def test_calls_parameter(self):
        self.validate_snippet(self.get_snippet_path("calls_parameter"))

    def test_return_call(self):
        self.validate_snippet(self.get_snippet_path("return_call"))

    def test_call(self):
        self.validate_snippet(self.get_snippet_path("call"))

    def test_parameter_call(self):
        self.validate_snippet(self.get_snippet_path("parameter_call"))

    def test_chained_calls(self):
        self.validate_snippet(self.get_snippet_path("chained_calls"))
